'''Exception'''


class WMError(Exception):
    '''Watermark Error'''
    # pass


class WMInvalidArgumentError(WMError):
    '''Invalid Argument Error'''

    def __init__(self, namespace: str, name: str):
        super().__init__()
        self._namespace = namespace
        self._name = name

    def __repr__(self):
        return f'Can not found "{self._name}" in {self._namespace}'

    def __str__(self):
        return self.__repr__()



'''module docstring'''

from textwatermark.defines import WMMethod

# 同形异义-字母
CONFUSABLES_CHARS = {
    'A': 'ΑАᎪꓮ𝐀𝖠𝗔',
    'B': 'ΒВᏴꓐ𝐁𝖡𝗕',
    'C': 'ⅭＣСᏟⲤꓚϹ',
    'D': 'ⅮᗞᗪᎠꓓ𝐃𝖣',
    'E': 'ⴹΕЕᎬꓰ𝐄𝖤',
    'F': 'ᖴϜꓝ𝐅𝖥𝙵𝟊',
    'G': 'Ꮐ𝙶ꓖ𝐆𝖦𝗚𝐺',
    'H': 'ᕼΗНᎻⲎꓧ𝖧',
    'I': 'ⅠΙІӀ𝐈𝙸𝚰',
    'J': 'ᒍЈᎫ𝐉𝙹𝖩𝗝',
    'K': 'KΚКᏦⲔꓗ𝐊',
    'L': 'ⅬᒪᏞⳐꓡ𝐋𝙻',
    'M': 'ⅯＭΜМᎷꓟ𝐌',
    'N': 'ＮΝⲚꓠ𝐍𝖭𝗡',
    'O': 'ଠⵔՕⲞꓳ𝚶𝐎',
    'P': 'ΡРᏢⲢꓑ𝐏𝖯',
    'Q': '𝐐𝑄𝑸𝖰𝗤𝘘𝙌',
    'R': 'ᎡᏒꓣ𝐑𝖱𝗥𝚁',
    'S': 'ЅՏᏚꓢ𝖲𝗦𝚂',
    'T': 'ΤТᎢꓔ𝖳𝗧𝚃',
    'U': 'Սꓴ𝚄𝐔𝖴𝗨𝘜',
    'V': 'ⅤᐯⴸᏙꓦ𝐕𝖵',
    'W': 'ԜᎳꓪ𝐖𝖶𝗪𝘞',
    'X': 'ΧХⲬⵝꓫ𐊐𐊴',
    'Y': 'ΥҮⲨꓬ𝐘𝖸𝗬',
    'Z': 'ΖᏃꓜꛉ𐕓𑢩Ⲍ',
    'a': 'аɑαᥲ⍺𐓟ꭤ',
    'b': 'ᖯЬᏏ𝐛𝖻𝗯𝚋',
    'c': 'ϲсᥴᴄⅽⲥ𐔛',
    'd': 'ⅾᑯꓒ𝐝𝖽𝗱𝙙',
    'e': 'еℯ𝐞𝖾𝗲𝘦𝙚',
    'f': 'ẝ𝐟𝕗𝖿𝗳𝙛𝘧',
    'g': 'ɡց𝐠𝑔𝒈𝗀𝗴',
    'h': 'һᏂ𝐡𝕙𝗁𝗵𝚑',
    'i': 'іᎥ𝐢𝕚𝗂𝗶𝚒',
    'j': 'ј𝚓ϳ𝐣𝕛𝗃𝗷',
    'k': '𝐤𝗄𝗸𝙠𝚔𝑘𝒌',
    'l': '𝑙𝒍𝓁𝓵𝗅𝗹𝙡',
    'm': 'ⅿ𝐦𝗆𝗺𝘮𝙢𝑚',
    'n': 'ո𝐧𝗇𝗻𝚗𝘯𝙣',
    'o': 'οоᴏⲟ𐓪օ௦',
    'p': 'рⲣ𝐩𝗉𝗽𝚙𝑝',
    'q': 'ԛգզ𝐪𝗊𝗾𝚚',
    'r': '𝚛𝐫𝒓𝓇𝓻𝗿𝗋',
    's': '᥉ѕꮪꜱ𝚜𝐬𝗌',
    't': '𝚝𝑡𝓉𝓽𝗍𝘵𝙩',
    'u': 'ᴜս𝐮𝗎𝘂𝚞𝑢',
    'v': 'ᴠⅴνѵ𝐯∨𝝼',
    'w': 'ᴡԝ𝐰𝗐𝘄𝘸𝙬',
    'x': 'ⅹх𝐱᙮𝗑𝘅𝘹',
    'y': 'уꭹ𝐲𝒚𝗒𝘆𝚢',
    'z': 'ᴢ𝐳𝑧𝗓𝘇𝘻𝙯',
}

for ikey, ival in CONFUSABLES_CHARS.items():
    CONFUSABLES_CHARS[ikey] = ikey + ival
method = WMMethod.FIND_AND_REPLACE



'''TextWatermark's Template'''

from textwatermark.defines import WMMethod
from textwatermark.templates import (combining_chars, invisible_chars,
                                     whitespace_chars)

CONFUSABLES_CHARS_LIST = combining_chars.CONFUSABLES_CHARS + \
    invisible_chars.CONFUSABLES_CHARS + whitespace_chars.CONFUSABLES_CHARS

CONFUSABLES_CHARS = dict(zip(CONFUSABLES_CHARS_LIST, CONFUSABLES_CHARS_LIST))

method = WMMethod.APPEND_AS_BINARY



'''Defines'''

from enum import Enum, IntEnum, unique


@unique
class WMMode(IntEnum):
    '''An enumerarion of watermark mode'''

    REAL_NUMBER = 1
    '''Real numbers, which will be represented directly by binary values, 
        with the smallest length'''

    LETTERS_LOWER_CASE = 2
    '''Lowercase letters, each letter is represented by 5 binary bytes'''

    LETTERS_UPPER_CASE = 3
    '''Uppercase letters, each letter is represented by 5 binary bytes'''

    LETTERS_MIXED_CASE = 4
    '''Mixed case letters, each character is represented by 6 binary bytes'''

    ALPHA_NUMERICAL = 5
    '''Uppercase and lowercase letters + numbers, each character is represented by 6 binary bytes'''

    ALPHA_NUMERICAL_SPECIAL = 6
    '''Uppercase and lowercase letters + numbers + special symbols, 
        each character is represented by 7 binary bytes'''

    UNICODE = 7
    '''Arbitrary Unicode characters, expressed in UTF-8 encoding'''


@unique
class WMMethod(IntEnum):
    '''Watermark insert methods'''

    # 查找特定字符并替换成水印字符
    FIND_AND_REPLACE = 1
    '''Find specific characters and replace with watermark characters'''

    # 在指定位置插入不可见水印
    INSERT_INTO_POSITION = 2
    '''Insert an invisible watermark at the specified position'''

    # 对文本中的字符进行修饰
    DECORATE_EACH_CHAR = 3
    '''Decorate characters in text'''

    # 在字符后添加单个水印
    APPEND_TO_CHAR = 4
    '''Add one watermark byte after one character'''

    # 以二进制形式在字符后面添加水印字符
    APPEND_AS_BINARY = 5
    '''Add one watermark byte after one character, in binary'''


@unique
class WhiteSpaceChars(Enum):
    '''Defines of white space chars

    From: <https://jkorpela.fi/chars/spaces.html>
    '''

    SPACE_1_4 = '\u0020'
    '''Depends on font, typically 1/4 em, often adjusted'''

    NO_BREAK_SPACE_1_4 = '\u00A0'
    '''As a space, but often not adjusted'''

    EN_QUAD_1_2 = '\u2000'
    '''1 en (= 1/2 em)'''

    EM_QUAD_1 = '\u2001'
    '''	1 em (nominally, the height of the font)'''

    EN_SPACE_1_2 = '\u2002'
    '''1 en (= 1/2 em)'''

    EM_SPACE_1 = '\u2003'
    '''1 em'''

    THREE_PER_EM_SPACE_1_3 = '\u2004'
    '''1/3 em'''

    FOUR_PER_EM_SPACE_1_4 = '\u2005'
    '''1/4 em'''

    SIX_PER_EM_SPACE_1_6 = '\u2006'
    '''1/6 em'''

    FIGURE_SPACE_LIKE_DIGITS = '\u2007'
    '''“Tabular width”, the width of digits'''

    PUNCTUATION_SPACE_LIKE_DOT = '\u2008'
    '''The width of a period “.”'''

    THIN_SPACE_1_6 = '\u2009'
    '''	1/5 em (or sometimes 1/6 em)'''

    HAIR_SPACE_1_8 = '\u200A'
    '''Narrower than `THIN SPACE`'''

    NARROW_NO_BREAK_SPACE_1_6 = '\u202F'
    '''Narrower than `NO-BREAK SPACE` (or `SPACE`), 
    “typically the width of a thin space or a mid space”'''

    MEDIUM_MATHEMATICAL_SPACE_2_9 = '\u205F'
    '''4/18 em'''



'''watermark template of homograph numbers'''

from textwatermark.defines import WMMethod

# 同形异义-数字
CONFUSABLES_CHARS = {
    '0': '᱐𝟘𝟎𝟢𝟬𝟶',
    '1': 'Ӏ𝟙𝟏𝟣𝟭𝟷',
    '2': 'ᒿ𝟚𝟐𝟤𝟮𝟸',
    '3': 'Ⳍ𝟛𝟑𝟥𝟯𝟹',
    '4': 'Ꮞ𝟜𝟒𝟦𝟰𝟺',
    '5': 'Ƽ𝟝𝟓𝟧𝟱𝟻',
    '6': 'Ⳓ𝟞𝟔𝟨𝟲𝟼',
    '7': 'ገ𝟟𝟕𝟩𝟳𝟽',
    '8': '৪𝟠𝟖𝟪𝟴𝟾',
    '9': 'Ꝯ𝟡𝟗𝟫𝟵𝟿',
}

for ikey, ival in CONFUSABLES_CHARS.items():
    CONFUSABLES_CHARS[ikey] = ikey + ival
method = WMMethod.FIND_AND_REPLACE



'''
WMTemplate is mainly used to load template, set related parameters, 
and insert watermark into text
'''
# pylint: disable=consider-using-enumerate


import re
from typing import Union

import numpy as np

from textwatermark.defines import WMMethod


class WMTemplate:
    '''Class of WMTemplate

    Example
    --------
    ```py 
    wmt = WMTemplate(
        template.confusables_chars,
        template.confusables_chars_length,
        template.method,
        confusables_chars_key)

    wm_text = wmt.insert_watermark(
        text=text,
        wm_final=wm_final,
        start_at=start_at,
        loop=False)

    wm_final = wmt.retrieve_watermark(
        wm_text=wm_text,
        wm_base=wmt.wm_base,
        wm_len=len(wm_final),
        start_at=start_at)
    ```
    '''

    def __init__(self, confusables_chars: Union[dict, list], wm_base: int, method: WMMethod,
                 confusables_chars_key: str = ''):
        '''Init of WMTemplate

        Args:
            confusables_chars (Union[dict, list]): The default template is in templates, 
                you can your the modified custom template 
            wm_base (int): Watermark convert base. Must be less than 
                or equal to confusables_chars_length 
            method (WMMethod):  Watermark insertion method, 
                see defines.WMMethod for details 
            confusables_chars_key (str): Key must be set when
                WMMethod is DECORATE_EACH_CHAR

        Note: Note of wm_base
            The value of wm_base will affect the range of watermark characters taken 
                from confusables_chars

        '''

        self.confusables_chars = confusables_chars
        '''The default template is in templates, 
                you can your the modified custom template '''
        if wm_base == 0:
            self.wm_base = WMTemplate.get_wm_base_from_tpl(
                self.confusables_chars)
            '''Watermark convert base. Must be less than or equal to confusables_chars_length '''
        else:
            self.wm_base = wm_base

        self.method = method
        '''Watermark insertion method, see defines.WMMethod for details '''

        if self.method in [WMMethod.DECORATE_EACH_CHAR, WMMethod.APPEND_AS_BINARY] and \
                confusables_chars_key not in self.confusables_chars:
            raise ValueError(
                f'When WMMethod in [DECORATE_EACH_CHAR,APPEND_AS_BINARY], confusables_chars_key: \
                    {confusables_chars_key} must can be found in confusables_chars\'s keys.')

        self.confusables_chars_key = confusables_chars_key
        '''Key must be set when WMMethod is DECORATE_EACH_CHAR'''

    @ staticmethod
    def get_wm_base_from_tpl(confusables_chars):
        '''get wm_base from template's confusables chars

        Args:
            confusables_chars (dict): confusables chars template

        Returns:
            (int): wm_base

        '''
        if isinstance(confusables_chars, list):
            wm_base = len(confusables_chars)
        elif isinstance(confusables_chars, dict):
            wm_base = len(list(confusables_chars.values())[0])
        else:
            raise TypeError('confusables_chars must be list or dict')

        wm_base = max(wm_base, 2)
        wm_base = min(wm_base, 36)

        return wm_base
    ################################# Clean watermark#################################

    def clean_html_tags(self, html):
        '''Clean html tags from watermark.

        Args:
            html (str): HTML code

        Returns:
            (str): Cleaned HTML code
        '''
        return re.sub(r'<[^<]+?>', '', html)

    def clean_text(self, text: str):
        '''clean confusables chars from text

        Args:
            text (str): Text to clean

        Returns:
            (str): Cleaned text

        Raises:
            ValueError: If unknown method

        '''
        text = self.clean_html_tags(text)

        match self.method:
            case WMMethod.FIND_AND_REPLACE:
                for key, vals in self.confusables_chars.items():
                    if isinstance(vals, list):
                        vals = ''.join(vals)
                    vals = vals.replace(key, '')
                    for val in vals:
                        text = text.replace(val, key)
            case WMMethod.DECORATE_EACH_CHAR:
                # use self.clean_html_tags by default
                pass
            case WMMethod.INSERT_INTO_POSITION | WMMethod.APPEND_TO_CHAR:
                for val in self.confusables_chars:
                    text = text.replace(val, '')
            case WMMethod.APPEND_AS_BINARY:
                text = text.replace(
                    self.confusables_chars[self.confusables_chars_key], '')
            case _:
                raise ValueError(f'Unknown WMMethod: {self.method}')
        # print(text)
        return text

    ################################# Insert watermark#################################

    def insert_watermark(self, text: str, wm_final: str, start_at: int = 0, loop: bool = False):
        '''Insert watermark into text

        Args:
            text (str): Text to be watermarked
            wm_final (str): The final watermark string
            start_at (int): The offset position for insert watermark
            loop (bool): Whether to insert the watermark in a loop

        Returns:
            (str): Watermarked text

        Raises:
            ValueError: If start_at is larger than text length
            ValueError: If there is not enough space to insert a watermark
            VaueError: If unknown watermark method

        '''
        if start_at > len(text):
            raise ValueError(
                f'Start_at is {start_at}, larger than text length: {len(text)}')

        text = self.clean_text(text=text)

        if start_at > 0:
            out_text = text[0:start_at]
            text = text[start_at:]
        else:
            out_text = ''

        match self.method:
            case WMMethod.FIND_AND_REPLACE:
                times = self.check_find_and_replace_space(
                    text, len(wm_final))
                if times < 1:
                    raise ValueError(
                        f'There is not enough space to insert a watermark: {wm_final}')

                out_text += self._find_and_replace(
                    text=text, wm_final=wm_final, loop=loop)
            case WMMethod.DECORATE_EACH_CHAR:
                if len(wm_final) > len(text):
                    raise ValueError(
                        f'There is not enough space to insert a watermark: {wm_final}')

                out_text += self._decorate_each_char(
                    text=text, wm_final=wm_final, loop=loop)
            case WMMethod.INSERT_INTO_POSITION:
                out_text += self._insert_into_position(
                    text=text, wm_final=wm_final)
            case WMMethod.APPEND_TO_CHAR:
                if len(wm_final) > len(text):
                    raise ValueError(
                        f'There is not enough space to insert a watermark: {wm_final}')

                out_text += self._append_to_char(
                    text=text, wm_final=wm_final, loop=loop)
            case WMMethod.APPEND_AS_BINARY:
                if len(wm_final) > len(text):
                    raise ValueError(
                        f'There is not enough space to insert a watermark: {wm_final}')

                out_text += self._append_as_binary(
                    text=text, wm_final=wm_final, loop=loop)
            case _:
                raise ValueError(f'Unknown watermark method: {self.method}')

        return out_text

    def check_find_and_replace_space(self, text: str, wm_len: int):
        '''Check if enough space for insert watermark

        Args:
            text (str): text to check
            wm_len (int): watermark length
        Returns:
            (int): number of times can insert watermark

        Note: Why use this method?
            Because WMMethod.FIND_AND_REPLACE mode must need enough space 
            to find special char and replace to watermarked char
        '''
        actual_len = 0
        confusables_chars_keys = ''.join(self.confusables_chars.keys())
        # 计算可插入的水印空间大小
        for char in text:
            if confusables_chars_keys.find(char) != -1:
                actual_len += 1
        times = actual_len // wm_len
        return times

    def _find_and_replace(self, text: str, wm_final: str, loop: bool):
        confusables_chars_keys = ''.join(self.confusables_chars.keys())
        wm_idx = 0
        wm_text = ''
        for idx in range(len(text)):
            if not loop and wm_idx >= len(wm_final):
                wm_text += text[idx:]
                break
            # 查找并替换成水印字符
            if confusables_chars_keys.find(text[idx]) != -1:
                offset = wm_final[wm_idx % len(wm_final)]
                wm_text += self.confusables_chars[text[idx]][int(offset)]
                wm_idx += 1
            else:
                wm_text += text[idx]
        return wm_text

    def _decorate_each_char(self, text: str, wm_final: str, loop: bool):
        wm_idx = 0
        wm_text = ''
        for idx in range(len(text)):
            if not loop and wm_idx >= len(wm_final):
                wm_text += text[idx:]
                break
            offset = wm_final[wm_idx % len(wm_final)]
            wm_text += self.confusables_chars[self.confusables_chars_key][int(
                offset)].replace('{char}', text[idx])

            wm_idx += 1

        return wm_text

    def _insert_into_position(self, text: str, wm_final: str):
        wm_out_str = ''
        for wm_char in wm_final:
            wm_out_str += self.confusables_chars[int(
                wm_char, self.wm_base)]
        return wm_out_str + text

    def _append_to_char(self, text: str, wm_final: str, loop: bool):
        wm_idx = 0
        wm_text = ''
        for idx in range(len(text)):
            if not loop and wm_idx >= len(wm_final):
                wm_text += text[idx:]
                break
            # 在每个字符串后添加字符
            offset = wm_final[wm_idx % len(wm_final)]
            wm_text += text[idx] + \
                self.confusables_chars[int(str(offset), self.wm_base)]
            wm_idx += 1
        return wm_text

    def _append_as_binary(self, text: str, wm_final: str, loop: bool):
        wm_idx = 0
        wm_text = ''
        for idx in range(len(text)):
            if not loop and wm_idx >= len(wm_final):
                wm_text += text[idx:]
                break
            # 在每个字符串后添加字符
            offset = wm_final[wm_idx % len(wm_final)]
            if offset == '0':
                wm_text += text[idx]
            else:
                wm_text += text[idx] + \
                    self.confusables_chars[self.confusables_chars_key]
            wm_idx += 1
        return wm_text

    ################################# Retrieve watermark#################################

    def retrieve_watermark(self, wm_text: str, wm_base: int, wm_len: int, start_at: int = 0):
        '''Retrieve watermark from watermarked text

        Args:
            wm_text (str): Text which has be watermarked
            wm_base (int): The encoding base of the watermark string
            wm_len (int): The fixed length of watermark
            start_at (int): The offset position of insert watermark

        Returns:
            (str): The final watermark retrieved from text

        Raises:
            ValueError: If `start_at` is larger than `wm_text` length
            ValueError: If the length of retrieve watermark is not larger then `wm_len`
            ValueError: If unknown watermark method

        '''
        if start_at > len(wm_text):
            raise ValueError(
                f'start_at: {start_at} is larger than wm_text length: {len(wm_text)}')

        if start_at > 0:
            wm_text = wm_text[start_at:]

        wm_str = ''
        match self.method:
            case WMMethod.FIND_AND_REPLACE:
                wm_str = self._retrieve_find_and_replace(
                    wm_text, wm_len)
            case WMMethod.DECORATE_EACH_CHAR:
                wm_str = self._retrieve_decorate_each_char(
                    wm_text, wm_base, wm_len)
            case WMMethod.INSERT_INTO_POSITION:
                wm_str = self._retrieve_insert_into_position(
                    wm_text, wm_base, wm_len)
            case WMMethod.APPEND_TO_CHAR:
                wm_str = self._retrieve_append_to_char(
                    wm_text, wm_base, wm_len)
            case WMMethod.APPEND_AS_BINARY:
                wm_str = self._retrieve_append_as_binary(
                    wm_text, wm_len)
            case _:
                raise ValueError(f'Unknown watermark method: {self.method}')

        if len(wm_str) < wm_len:
            raise ValueError(
                f'Watermark is: {wm_str}, length is less than {wm_len}')

        return wm_str[0: wm_len]

    def _retrieve_find_and_replace(self, wm_text: str, wm_len: int):
        vals = self.confusables_chars.values()
        confusables_chars_length = len(list(vals)[0])
        confusables_chars_values = ''
        for val in vals:
            if isinstance(val, list):
                confusables_chars_values += ''.join(val)
            if isinstance(val, str):
                confusables_chars_values += val
        wm_temp = ''
        for char in wm_text:
            distances = confusables_chars_values.find(char)
            if distances != -1:
                offset = int(distances % confusables_chars_length)
                # print(distances, offset)
                wm_temp += str(offset)
            if len(wm_temp) > wm_len:
                break
        return wm_temp

    def _retrieve_decorate_each_char(self, wm_text: str, wm_base: int, wm_len: int):
        confusables_chars = self.confusables_chars[self.confusables_chars_key]
        for idx in range(len(confusables_chars)):
            offset = int(idx % wm_base)
            regex = confusables_chars[idx].replace('{char}', '.*?')
            wm_text = re.sub(regex, f'<wm_char>{offset}</wm_char>', wm_text)

        matched = re.findall(r'<wm_char>(.*?)</wm_char>', wm_text)
        return ''.join(matched)[0:wm_len]

    def _retrieve_insert_into_position(self, wm_text: str, wm_base: int, wm_len: int):
        for idx in range(len(self.confusables_chars)):
            wm_val = np.base_repr(idx, wm_base)
            wm_text = wm_text.replace(
                self.confusables_chars[idx], f'<wm_char>{wm_val}</wm_char>')

        matched = re.findall(r'<wm_char>(.*?)</wm_char>', wm_text)
        return ''.join(matched)[0:wm_len]

    def _retrieve_append_to_char(self, wm_text: str, wm_base: int, wm_len: int):
        '''retrieve watermark from append to char'''
        return self._retrieve_insert_into_position(wm_text, wm_base, wm_len)

    def _retrieve_append_as_binary(self, wm_text: str, wm_len: int):
        '''retrieve watermark from append char as binary'''
        confusables_char = self.confusables_chars[self.confusables_chars_key]
        wm_start = False
        wm_final = ''

        for idx, char in enumerate(wm_text):
            if len(wm_final) == wm_len:
                break
            if char == confusables_char and wm_start is False:
                wm_start = True
            if wm_start is False:
                continue

            if char == confusables_char:
                wm_final += '1'
                continue

            if char != confusables_char and \
                    (idx+1 >= len(wm_text) or wm_text[idx+1] != confusables_char):
                wm_final += '0'
                continue
        if len(wm_final) != wm_len:
            raise ValueError(
                f'Retrieved watermark {wm_final}, length is not equal to {wm_len}')

        return wm_final



'''module docstring'''

from textwatermark.defines import WMMethod

# 空格字符
CONFUSABLES_CHARS = [
    '<a></a>',
    '<abbr></abbr>',
    '<b></b>',
    '<bdo></bdo>',
    '<big></big>',
    '<cite></cite>',
    '<code></code>',
    '<dfn></dfn>',
    '<em></em>',
    '<i></i>',
    '<kbd></kbd>',
    '<var></var>',
    '<samp></samp>',
    '<span></span>',
    '<strong></strong>',
    '<small></small>',
    '<sub></sub>',
    '<sup></sup>',
    '<tt></tt>',
    '<u></u>',
]

method = WMMethod.INSERT_INTO_POSITION



'''module docstring'''

from textwatermark.common import add_head_n_tial_to_dict
from textwatermark.defines import WMMethod

# 组合字符，如音标等
CONFUSABLES_CHARS_START_WITH = '<span style="'
CONFUSABLES_CHARS_END_WITH = ';">{char}</span>'
CONFUSABLES_CHARS = {
    'text-decoration-underline': ['text-decoration:none', 'text-decoration:underline'],
    'text-decoration-line-through': ['text-decoration: none', 'text-decoration: line-through'],
    'text-decoration-overline': ['text-decoration: none', 'text-decoration: overline']
}

add_head_n_tial_to_dict(
    CONFUSABLES_CHARS, CONFUSABLES_CHARS_START_WITH, CONFUSABLES_CHARS_END_WITH)

method = WMMethod.DECORATE_EACH_CHAR



'''module docstring'''

from textwatermark.defines import WhiteSpaceChars, WMMethod

# 空格字符
CONFUSABLES_CHARS = {
    '\u0020': [
        WhiteSpaceChars.SPACE_1_4.value,
        # WhiteSpaceChars.NO_BREAK_SPACE_1_4.value,
        # WhiteSpaceChars.EN_QUAD_1_2.value,
        # WhiteSpaceChars.EM_QUAD_1.value,
        WhiteSpaceChars.EN_SPACE_1_2.value,
        WhiteSpaceChars.EM_SPACE_1.value,
        # WhiteSpaceChars.THREE_PER_EM_SPACE_1_3.value,
        # WhiteSpaceChars.FOUR_PER_EM_SPACE_1_4.value,
        # WhiteSpaceChars.SIX_PER_EM_SPACE_1_6.value,
        # WhiteSpaceChars.FIGURE_SPACE_LIKE_DIGITS.value,
        # WhiteSpaceChars.PUNCTUATION_SPACE_LIKE_DOT.value,
        # WhiteSpaceChars.THIN_SPACE_1_6.value,
        WhiteSpaceChars.HAIR_SPACE_1_8.value,
        # WhiteSpaceChars.NARROW_NO_BREAK_SPACE_1_6.value,
        # WhiteSpaceChars.MEDIUM_MATHEMATICAL_SPACE_2_9.value,
    ]
}

method = WMMethod.FIND_AND_REPLACE



'''module docstring'''

from textwatermark.common import add_head_n_tial_to_dict
from textwatermark.defines import WMMethod

# 组合字符，如音标等
CONFUSABLES_CHARS_START_WITH = '<span style="color: '
CONFUSABLES_CHARS_END_WITH = '">{char}</span>'
CONFUSABLES_CHARS = {
    'black1': ['#000', '#111'],
    'black2': ['#000', '#222'],
    'black3': ['#000', '#333'],
    'black4': ['#000', '#444'],
    'white1': ['#fff', '#eee'],
    'white2': ['#fff', '#ddd'],
    'white3': ['#fff', '#ccc'],
    'white4': ['#fff', '#bbb'],
}

add_head_n_tial_to_dict(
    CONFUSABLES_CHARS, CONFUSABLES_CHARS_START_WITH, CONFUSABLES_CHARS_END_WITH)

method = WMMethod.DECORATE_EACH_CHAR



'''Main Class of TextWatermark'''
# pylint: disable=too-many-instance-attributes

import json
import os
import sys
from typing import Union

from textwatermark import __version__
from textwatermark.conversion import WMConversion
from textwatermark.defines import WMMethod, WMMode
from textwatermark.template import WMTemplate
from textwatermark.template_type import WMTemplateType


class TextWatermark:
    '''
    A class to insert watermark into plain text

    Example:
    --------
    ```py
    import os

    from textwatermark.defines import WMMode
    from textwatermark.main import TextWatermark
    from textwatermark.templates import homograph_numbers

    wm_mode = WMMode.REAL_NUMBER
    template = homograph_numbers
    wm_max = '9'*9
    confusables_chars_key = ''

    # init
    wm = TextWatermark(wm_mode)


    wm = set_tpl(template.CONFUSABLES_CHARS,
                 template.method,
                 confusables_chars_key)
    wm.set_wm_max(wm_max)
    wm.set_text_file(os.path.abspath(
        os.path.dirname(__file__)+'/../tests/text/1.txt'))

    # Insert watermark to text
    wm_str = '123456789'

    wm_text = wm.insert_watermark(wm_str)

    # Save the parameters to retrieve the watermark
    params = wm.export_params()

    # retrieve the watermark
    wm_out_str = TextWatermark.retrieve_watermark(wm_text, params)
    print(wm_text)

    assert wm_out_str == wm_str
    ```

    '''

    text: str
    '''Text to be watermarked.

        Set text by `self.set_text(text)` or `self.set_text_file(file_path)`
    '''

    wmc: WMConversion
    '''Instance of WMConverison'''
    wmt: WMTemplate
    '''Instance of WMTemplate'''

    wm_max: str
    '''Maximum watermark string.'''

    wm_fixed_len: int = 0
    '''Watermark fixed(max length) length.'''

    tpl_type: str = ''
    '''template type in WMTemplateType'''

    def __init__(self, wm_mode: WMMode, wm_base: int = 0, start_at: int = 0, wm_loop: bool = False):
        '''Class TextWatermark init

        Args:
            wm_mode (WMMode): WMMode to be set
            wm_base (int): Watermark convert base.
            start_at (int, optional): index where the watermark will be inserted. Defaults to 0.
            wm_loop (bool, optional): If watermark will be repeated until it is inserted to end.
                Defaults to False.

        Raises:
            ValueError: if `base` is not in the range [2, 36]
        '''

        self.wm_mode = wm_mode
        '''Watermark encoding mode. Setting different watermark modes will effectively
        reduce the size of the watermark string.'''

        self.wm_base = wm_base
        '''watermark convert base must be between 2 and 36,
            and less or euqal than len(list({template}.CONFUSABLES_CHARS.values())[0])
        '''

        self.start_at = start_at
        '''index where the watermark will be inserted. Defaults to 0.'''
        self.wm_loop = wm_loop
        '''If watermark will be repeated until it is inserted to end.
                Defaults to False.'''

    def set_tpl(self, confusables_chars: Union[dict, list],
                method: WMMethod, confusables_chars_key: str = ''):
        '''Set template arguments and get instance of WMTemplate

        Args:
            confusables_chars (Union[dict, list]): Confusables chars
            method (WMMethod): Method of watermark
            confusables_chars_key (str): Confusables chars key

        '''

        self.wmt = WMTemplate(
            confusables_chars, self.wm_base, method, confusables_chars_key)

        if self.wm_base == 0:
            self.wm_base = self.wmt.wm_base

    def set_tpl_type(self, tpl_type: WMTemplateType, confusables_chars_key: str = ''):
        '''Set template type

        Args:
            tpl_type (WMTemplateType): Template type
            confusables_chars_key (str): Confusables chars key

        '''
        template = tpl_type.value

        self.wmt = WMTemplate(template.CONFUSABLES_CHARS, self.wm_base,
                              template.method, confusables_chars_key)

        if self.wm_base == 0:
            self.wm_base = self.wmt.wm_base

        self.tpl_type = tpl_type.name

    def set_wm_max(self, wm_max: str):
        '''Set the length of the longest (or largest) watermark string.

            Since the input watermark string is variable in length,
            this will bring trouble to the retrieval of the watermark:
            it will be difficult for us to determine the boundary of the watermark,
            or judge the loss of the watermark. Therefore, in this library,
            we will calculate and set the length of the longest (or largest)
            watermark string as the default length of the watermark string.

        Args:
            wm_max (str): The longest (or largest) watermark string.

        '''
        if not hasattr(self, 'wmt') or not self.wmt:
            raise ValueError(
                'Set up template with set_tpl or set_tpl_type first please.')

        self.wmc = WMConversion(self.wm_mode, self.wmt.wm_base)
        self.wm_max = wm_max
        self.wm_fixed_len = self.wmc.calc_max_wm_length(wm_max)

    def set_text(self, text: str):
        '''Set text string to watermark

        Args:
            text (str): text to be watermarked
        '''
        self.text = text

    def set_text_file(self, path: str):
        '''Set text string from read file.

        Args:
            path (str): Text file path

        Raises:
            ValueError: If file not found
            OSError: If file cannnot by read
        '''
        path = os.path.abspath(path)

        if not os.path.exists(path):
            raise ValueError(f'ERROR: file {path} does not exist')

        try:
            with open(path, 'r', encoding='utf-8') as file:
                self.set_text(file.read())
        except OSError as err:
            print(f'ERROR: cannot read file {path}, err is {err.strerror}')
            sys.exit()

    def insert_watermark(self,  wm_str: str):
        '''Insert watermark to text

        Args:
            wm_str (str): Watermark string

        Returns:
            (str): Watermarked text

        Raises:
            ValueError: If watermark string is larger than wm_max
            ValueError: If there is not enough space to insert a watermark
        '''
        if wm_str > self.wm_max:
            raise ValueError(
                f'ERROR: watermark:{wm_str} if larger than wm_max: {self.wm_max}')

        wm_final = self.wmc.wm_convert_to_arbitrary_base(wm_str)
        wm_final = '1' + wm_final.zfill(self.wm_fixed_len-1)
        if len(wm_final) > self.wm_fixed_len:
            raise ValueError(
                f'ERROR: watermark {wm_str} (convert to: {wm_final}) \
                    is too long: {len(wm_final)}, max length is {self.wm_fixed_len}')
        #
        wm_text = self.wmt.insert_watermark(
            text=self.text,
            wm_final=wm_final,
            start_at=self.start_at,
            loop=self.wm_loop)
        return wm_text

    def save_to_file(self, wm_text: str, path: str):
        '''Save watermarked text string to a file

        Args:
            wm_text (str): text to be watermarked
            path (str): save to text file path

        Raises:
            OSError: If file cannnot be created

        '''
        path = os.path.abspath(path)

        try:
            with open(path, 'w', encoding='utf-8') as file:
                file.write(wm_text)
        except OSError as err:
            print(f'ERROR: cannot write to file {path}, err is {err.strerror}')
            sys.exit()

    def export_params(self):
        '''Export watermark params to json string

        Returns:
            (str): Exported JSON string

        '''
        if self.tpl_type != '' and self.wmt.confusables_chars \
                == WMTemplateType[self.tpl_type].value.CONFUSABLES_CHARS:
            confusables_chars = []
        else:
            confusables_chars = self.wmt.confusables_chars

        params = {
            'tpl_type': self.tpl_type,
            'confusables_chars': confusables_chars,
            'confusables_chars_key': self.wmt.confusables_chars_key,
            'wm_base': self.wmt.wm_base,
            'method': self.wmt.method.value,
            'wm_mode': self.wm_mode,
            'wm_len': self.wm_fixed_len,
            'wm_loop': self.wm_loop,
            'start_at': self.start_at,
            'version': __version__,
        }
        # print(params)

        return json.dumps(params, ensure_ascii=False)

    @ staticmethod
    def retrieve_watermark(wm_text: str, params: json):
        '''Retrieve watermark from watermarked text
        Note: This is a static method
            You can call this method by `TextWatermark.retrieve_watermark`

        Warning: Retrieve watermark method is not a silver bullet
            In many cases, the watermarked text may be transferred several times,
            which may cause the watermark to change or disappear. At this time,
            the watermark information cannot be retrieved through this function.
            In the case of screen capture and screen capture, manual judgment is
            required to find out the watermark information.

        Args:
            wm_text (str): watermarked text
            params (json): params containing the watermark options

        Returns:
            (str): Watermark string.

        '''
        params = json.loads(params)
        # print(params)
        wm_len = params['wm_len']

        wmc = WMConversion(params['wm_mode'], params['wm_base'])

        if not params['confusables_chars']:
            tpl_type = params['tpl_type']
            if tpl_type not in WMTemplateType.__members__:
                raise ValueError(f'Invalid WMTemplateType: {tpl_type}')
            params['confusables_chars'] = WMTemplateType[tpl_type].value.CONFUSABLES_CHARS

        wmt = WMTemplate(
            params['confusables_chars'],
            params['wm_base'],
            params['method'],
            params['confusables_chars_key'])

        wm_out = wmt.retrieve_watermark(
            wm_text=wm_text,
            wm_base=wmt.wm_base,
            wm_len=wm_len,
            start_at=params['start_at'])

        wm_temp = wm_out[1:wm_len]
        wm_temp = wm_temp.lstrip('0')
        wm_out_str = wmc.wm_restore_from_arbitrary_base(wm_temp)
        return wm_out_str



'''module docstring'''

from textwatermark.common import add_head_n_tial_to_dict
from textwatermark.defines import WMMethod

# 组合字符，如音标等
CONFUSABLES_CHARS_START_WITH = '<span style="display: inline-block;position: relative;'
CONFUSABLES_CHARS_END_WITH = ';">{char}</span>'
CONFUSABLES_CHARS = {
    'up1': ['top:0', 'top:-1px'],
    'up2': ['top:0', 'top:-2px'],
    'up3': ['top:0', 'top:-3px'],
    'bottom1': ['bottom:0', 'bottom:-1px'],
    'buttom2': ['bottom:0', 'bottom:-2px'],
    'bottom3': ['bottom:0', 'bottom:-3px'],
}

add_head_n_tial_to_dict(
    CONFUSABLES_CHARS, CONFUSABLES_CHARS_START_WITH, CONFUSABLES_CHARS_END_WITH)

method = WMMethod.DECORATE_EACH_CHAR



'''module docstring'''

from textwatermark.common import add_head_n_tial_to_dict
from textwatermark.defines import WMMethod

# 组合字符，如音标等
CONFUSABLES_CHARS_START_WITH = '<span style="font-weight: '
CONFUSABLES_CHARS_END_WITH = ';">{char}</span>'
CONFUSABLES_CHARS = {
    'lighter': ['normal', 'lighter'],
    'bolder': ['normal', 'bolder'],
    '100-200': ['100', '200'],
    '200-300': ['200', '300'],
    '300-400': ['300', '400'],
    '400-500': ['400', '500'],
    '500-600': ['500', '600'],
    '600-700': ['600', '700'],
    '700-800': ['700', '800'],
    '800-900': ['800', '900'],
}

add_head_n_tial_to_dict(
    CONFUSABLES_CHARS, CONFUSABLES_CHARS_START_WITH, CONFUSABLES_CHARS_END_WITH)

method = WMMethod.DECORATE_EACH_CHAR



'''Template type define'''
from enum import Enum, unique

from textwatermark.templates import (binary_representation, combining_chars,
                                     font_color, font_float, font_size,
                                     font_style, font_weight,
                                     homograph_chinese, homograph_letters,
                                     homograph_numbers, homograph_punctuations,
                                     html_empty_tags, invisible_chars,
                                     space_chars, whitespace_chars)


@unique
class WMTemplateType(Enum):
    '''watermark template types'''

# WMMethod.INSERT_INTO_POSITION
    # 不可见字符
    INVISIBLE_CHARS = invisible_chars
    '''Invisible characters'''

    # 空白字符
    WHITESPACE_CHARS = whitespace_chars
    '''White space characters'''

    # HTML空白标签
    HTML_EMPTY_TAGS = html_empty_tags
    '''HTML empty tags'''


# WMMethod.FIND_AND_REPLACE
    SPACE_CHARS = space_chars
    '''Space characters'''

    # 同形异义字-数字
    HOMOGRAPH_NUMBERS = homograph_numbers
    '''Homograph numbers'''

    # 同形异义字-英文
    HOMOGRAPH_LETTERS = homograph_letters
    '''Homograph letters'''

    # 同形异义字-符号
    HOMOGRAPH_PUNCTUATIONS = homograph_punctuations
    '''Homograph punctuations'''

    # 同形异义字-汉字
    HOMOGRAPH_CHINESE = homograph_chinese
    '''Homograph Chinese'''

    # 繁体字
    # TRADITIONAL_CHINESE = 3007
    # 错别字-汉字
    # WRONG_CHINESE = 3008
    # 错别字-单词
    # WRONG_WORDS = 3009
    # 同义替换-数字
    # SYNONYM_NUMBERS = 3010
    # 同义替换-汉字
    # SYNONYM_CHINESE = 3011
    # 同义替换-单词
    # SYNONYM_WORDS = 3012


# WMMethod.APPEND_TO_CHAR
    # 组合字符
    COMBINING_CHARS = combining_chars
    '''Combining characters'''


# WMMethod.DECORATE_EACH_CHAR
    # 字体颜色
    FONT_COLOR = font_color
    '''Font color'''

    # 字体大小
    FONT_SIZE = font_size
    '''Font size'''

    # 字体粗细
    FONT_WEIGHT = font_weight
    '''Font weight'''

    # 字体浮动
    FONT_FLOAT = font_float
    '''Font float'''

    # 字体样式
    FONT_STYLE = font_style
    '''Font style'''

# WMMethod.APPEND_AS_BINARY

    BINARY_REPRESENTATION = binary_representation
    # '''Delete character'''



'''watermark template class of invisible chars'''

from textwatermark.defines import WMMethod

CONFUSABLES_CHARS = [
    '\u034f', '\u061c', '\u180b', '\u180c', '\u180d', '\u180e', '\u200b', '\u200e', '\u200f',
    '\u202a', '\u202b', '\u202c', '\u202d', '\ufe00', '\ufe01', '\ufe02', '\ufe03', '\ufe04',
    '\ufe05', '\ufe06', '\ufe07', '\ufe08', '\ufe09', '\ufe0a', '\ufe0b', '\ufe0c', '\ufe0d',
    '\ufe0e', '\ufe0f'
]

method = WMMethod.INSERT_INTO_POSITION
